/*

  yyynetwork - a networking library
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef __YYYNETWORK_HPP
#define __YYYNETWORK_HPP

#include <list>
#include <sigc++/signal_system.h>

#include "yyylowlevelnet.hpp"
#include "binarydefs.hpp"
#include "fatal.hpp"
#include "yyysmartptr.hpp"
#include "yyysigdeferer.hpp"
#include "yyyexceptions.hpp"

namespace yyynetwork {

  template <class coopscheduler_, class binbuffer_> class server;
  template <class coopscheduler_, class binbuffer_> class client;

  class dns_hostname_to_ipaddress {
  public:
    // you can do multiple lookups at the same time.
    // ipv4address will be undefined if lookup failed.
    SigC::Signal3 <void, dns_hostname_to_ipaddress &, const hostname &, const ipv4address &> & lookup (const hostname &h);
  private:
  };

  // connection is a implementation of the binstream concept
  template <class coopscheduler_, class binbuffer_>
  class connection : public SigC::Object {
  public:
    typedef coopscheduler_ coopscheduler;
    typedef binbuffer_ binbuffer;
    enum direction_t { OUTGOING, INCOMING };
    enum state_t { CONNECTED, DISCONNECTED };

    binbuffer receivebuffer, sendbuffer;
    SigC::Signal1<void,connection&> receive;

    SigC::Signal1<void,connection&> sig_disconnected;
  private:
    coopscheduler *const mysched;
    typedef typename binbuffer::index index;
    static const index READBLOCKSIZE=4096;
    friend class server<coopscheduler,binbuffer>;
    friend class client<coopscheduler,binbuffer>;
    typedef typename coopscheduler::ionotify ionot_t;
    ipv4address foreign_ipv4address;
    portnumber foreign_portnumber;
    direction_t direction;
    state_t state;
    ionot_t ionot_read, ionot_write;
    tcphandle handle;
    void readslot(ionot_t &r)
    {
      index oldsize=receivebuffer.size();
      receivebuffer.resize(oldsize+READBLOCKSIZE);
      int realread=handle.read(&receivebuffer[oldsize],READBLOCKSIZE);
      if (realread>0) receivebuffer.resize(oldsize+index(realread));
      else receivebuffer.resize(oldsize);
      
      if (realread<=0) disconnect();
      else receive.emit(*this);
    };
    void writeslot(ionot_t &w)
    {
      index todo=sendbuffer.size();
      if (todo>READBLOCKSIZE) todo=READBLOCKSIZE;
      index realwrite=handle.write(&sendbuffer[0],todo);
      if (realwrite>0) {
	sendbuffer.erase(0,index(realwrite));
	if (sendbuffer.size()==0) ionot_write.setactive(false);
      } else disconnect();
    };
    void activate() // for client and server.
    {
      ionot_read.fire.connect(SigC::slot(*this,&connection<coopscheduler,binbuffer>::readslot));
      ionot_read.setfilehandle(handle.getfilehandle(),ionot_t::READ);
      ionot_write.fire.connect(SigC::slot(*this,&connection<coopscheduler,binbuffer>::writeslot));
      ionot_write.setfilehandle(handle.getfilehandle(),ionot_t::WRITE);
      ionot_write.setactive(false);
    };
  public:
    inline const direction_t getdirection () const { return direction; };
    inline const state_t getstate () const { return state; };
    inline const ipv4address & getforeign_ipv4address () const { return foreign_ipv4address; };
    inline const portnumber & getforeign_portnumber () const { return foreign_portnumber; };

    connection (coopscheduler *const mysched_) : mysched(mysched_), ionot_read(mysched), ionot_write(mysched) {};
    ~connection () {};
    // you cannot copy a connection
    connection (const connection &);
    connection & operator= (const connection &);

    void send()
    {
      if (state!=DISCONNECTED) {
	if (sendbuffer.size()) ionot_write.setactive(true);
      }else{
	sendbuffer.erase(0,sendbuffer.size());
      };
    };

    void disconnect()
    {
      if (state!=DISCONNECTED) {
	state=DISCONNECTED;
	ionot_read.setactive(false);
	ionot_write.setactive(false);
	handle.close();
	sendbuffer.erase(0,sendbuffer.size());
	receivebuffer.erase(0,receivebuffer.size());
	sig_disconnected.emit(*this);
      };
    };
  };

  template <class coopscheduler_, class binbuffer_>
  class server : public SigC::Object {
  public:
    except_mk(addressinuse, "local port is already in use");
    except_mk(accessdenied, "you don't have permission to bind that port");

    typedef coopscheduler_ coopscheduler;
    typedef binbuffer_ binbuffer;
    typedef yyysmartptr::smartptr<connection<coopscheduler, binbuffer> > sptr_connection_t;

  private:
    typedef typename coopscheduler::ionotify ionot;

  public:
    server (coopscheduler *const mysched_ = coopscheduler::gsched) : mysched(mysched_), io(mysched)
    {
      io.fire.connect(SigC::slot(*this,&server<coopscheduler,binbuffer>::work));
    };
    void setport (const portnumber &p) // start accepting new connections.
    {
      shutdown();
      acc.makesocket();
      acc.setnonblocking();
      acc.setreuse();
      tcphandle::bind_result_t bres;
      acc.bind(p,bres);
      switch (bres) {
      case tcphandle::BINDED: break;
      case tcphandle::ADDRESSINUSE: acc.close(); throw addressinuse();
      case tcphandle::ACCESSDENIED: acc.close(); throw accessdenied();
      };
      acc.listen();
      io.setfilehandle(acc.getfilehandle(),ionot::READ);
    };
    void shutdown () // stop accepting new connections.
    {
      io.setactive(false);
      acc.close();
    };

    SigC::Signal2<void, server &, sptr_connection_t & > resultavailable;

  private:
    tcphandle acc;
    coopscheduler *const mysched;
    ionot io;

    void work (ionot &)
    {
      sptr_connection_t s=new connection<coopscheduler,binbuffer>(mysched);
      tcphandle::accept_result_t res;
      acc.accept(s->handle,s->foreign_ipv4address,s->foreign_portnumber,res);
      switch (res) {
      case tcphandle::ACCEPTED: break;
      case tcphandle::WOULDBLOCK: fatal("server::work: tcphandle::accept returned WOULDBLOCK"); break;
      };
      s->handle.setsolinger();
      s->handle.setnonblocking();
      s->direction=connection<coopscheduler,binbuffer>::INCOMING;
      s->state=connection<coopscheduler,binbuffer>::CONNECTED;
      s->activate();
      resultavailable.emit(*this,s);
    };
  };

  template <class coopscheduler_, class binbuffer_>
  class client : public SigC::Object {
  public:
    typedef coopscheduler_ coopscheduler;
    typedef binbuffer_ binbuffer;
    typedef yyysmartptr::smartptr<connection<coopscheduler, binbuffer> > sptr_connection_t;

  private:
    typedef typename coopscheduler::ionotify ionot;

    class result {
      friend class yyynetwork::client<coopscheduler,binbuffer>;
    public:
      enum error_t { UNREACHABLE, REFUSED };
      inline const bool isconnected () const { return isconnected_rep; };
      const sptr_connection_t &getconnection () const // fatal if isconnected()==false
      { if (!isconnected()) fatal("client::result::getconnection: not connected"); return c; };
      const error_t geterror () const // fatal if isconnected()==true
      { if (isconnected()) fatal("client::result::geterror: i am connected!"); return err; };
    private:
      error_t err;
      sptr_connection_t c;
      bool isconnected_rep;
    };

    struct job;
    typedef yyysmartptr::smartptr<job> sptr_job_t;
    struct job {
      job (coopscheduler *const mysched) : not(mysched), sigdef(mysched) {};
      result res;
      ionot not;
      SigC::Signal2<void, client &, result &> resultavailable;
      yyysigdeferer::sigdeferer1<sptr_job_t&, coopscheduler> sigdef;
      tcphandle::connect_result_t conres;
    };
    list <sptr_job_t> jobs;

  public:
    client (coopscheduler *const mysched_ = coopscheduler::gsched) : mysched(mysched_) {};

    // the returned signal reference is only valid until the signal fired, so best not store it!
    SigC::Signal2<void,client &,result &> & netconnect (const ipv4address &foreign_ipv4address_,
							const portnumber &foreign_portnumber_)
    {
      sptr_job_t j=new job(mysched);
      jobs.push_back(j);
      j->res.c=new connection<coopscheduler,binbuffer>(mysched);
      j->res.c->foreign_ipv4address = foreign_ipv4address_;
      j->res.c->foreign_portnumber = foreign_portnumber_;
      j->sigdef.send.connect(SigC::slot(*this,&client<coopscheduler,binbuffer>::init));
      j->sigdef.receive(j);
      return j->resultavailable;
    };
    
  private:
    void init(sptr_job_t &j)
    {
      j->res.c->handle.makesocket();
      j->res.c->handle.setnonblocking();
      j->res.c->handle.setsolinger();
      j->res.c->handle.connect(j->res.c->foreign_ipv4address, j->res.c->foreign_portnumber, j->conres);
      if (j->conres==tcphandle::INPROGRESS) {
	j->not.fire.connect(SigC::slot(*this,&client<coopscheduler,binbuffer>::work));
	j->not.setfilehandle(j->res.c->handle.getfilehandle(),ionot::WRITE);
      }else done(j);
    };

    void work(ionot &i)
    {
      sptr_job_t j;
      list<sptr_job_t>::iterator it;
      for (it=jobs.begin();it!=jobs.end();++it) {
	if (&(*it)->not==&i) {
	  j=*it;
	  break;
	};
      };
      jobs.erase(it);
      j->res.c->handle.connect(j->res.c->foreign_ipv4address, j->res.c->foreign_portnumber, j->conres);
      done(j);
    };

    void done(sptr_job_t &j)
    {
      switch (j->conres) {
      case tcphandle::CONNECTED: {
	j->res.c->direction=connection<coopscheduler,binbuffer>::OUTGOING;
	j->res.c->state=connection<coopscheduler,binbuffer>::CONNECTED;
	j->res.isconnected_rep=true;
	j->res.c->activate();
	break;
      };
      case tcphandle::REFUSED: j->res.isconnected_rep=false; j->res.err=result::REFUSED; j->res.c.kill(); break;
      case tcphandle::INPROGRESS: fatal("client::done: tcphandle::connect returned INPROGRESS"); break;
      case tcphandle::UNREACHABLE: j->res.isconnected_rep=false; j->res.err=result::UNREACHABLE; j->res.c.kill(); break;
      };
      j->resultavailable.emit(*this,j->res);
    };

    coopscheduler *const mysched;
  };

};

#endif
